unit mKnob;
{ TmKnob : Marco Caselli's Knob Control rel. 1.0
 This component emulate the volume knob you could find on some HiFi devices;
**********************************************************************
* Feel free to use or give away this software as you see fit.        *
* Please leave the credits in place if you alter the source.         *
*                                                                    *
* This software is delivered to you "as is",                         *
* no guarantees of any kind.                                         *
*                                                                    *
* If you find any bugs, please let me know, I will try to fix them.  *
* If you modify the source code, please send me a copy               *
*                                                                    *
* If you like this component,  and also if you dislike it ;), please *
* send me an E-mail with your comment                                *
* Marco Caselli                                                      *
* Web site : http://members.tripod.com/dartclub                      *
* E-mail   : mcaselli@iname.com                                      *
*                                                                    *
* Thank to guy at news://marcocantu.public.italian.delphi            *
* for some math code. Check the site http://www.marcocantu.com       *
**********************************************************************
*** Sorry for my bad english ...............
 Properties :
      AllowUserDrag : Boolean; Specify if user can or not drag the control
                               to a new value using mouse;
      FaceColor : TColor;      Color of knob face;
      TickColor : TColor;      Color of tick mark;
      Position : Longint;      Current position of the knob;
      MarkStyle: TMarkStyle;   Specify style of the tick mark ( actually only
                               line or filled circle;
      RotationEffect:Boolean;    If True, the knob will shake emulating a rotation
                                  visual effect.
      Position:Longint;        Current value of knob;
      Max : Longint;           Upper limit value for Position;
      Min : Longint;           Lower limit value for Position;

 Events:
    property OnChange :        This event is triggered every time you change the
                               knob value;
*******************************************************************************}

interface

uses
   SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
   Forms, Dialogs, math,comctrls;

type
TKnobChangeEvent = procedure(Sender : TObject; Value : Longint) of object;
TMarkStyle=(msLine, msCircle); // More to come.......

TmKnob = class(TCustomControl)
private
   { Private declarations }
   fMaxValue 			: Longint;
   fMinValue 			: Longint;
   fCurValue 			: Longint;
   fColor 			: TColor;
   fTickColor 			: TColor;
   fDrag 			: Boolean;
   fKnobChangeEvent 	        : TKnobChangeEvent;
   FollowMouse 		        : Boolean;
   fMarkStyle                   : TMarkStyle;
   fRotationEffect                : Boolean;
   procedure SetMaxValue(Value : Longint);
   procedure SetCurValue(Value : Longint);
   procedure SetColor(Color : TColor);
   procedure SetTickColor(Value : TColor);
   procedure SetMarkStyle(Value:TMarkStyle);
   procedure SetDrag(Value : Boolean);
   procedure UpdatePosition(X, Y: Integer);
   function  ConstrainValue(Minimum, Maximum, Value : Longint) : Longint;

protected  { Protected declarations }
   procedure KnobChange;
   procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
   procedure Paint; override;

public
   { Public declarations }
   constructor Create(AOwner : TComponent); override;

published
   { Published declarations }
   property Align;
   property AllowUserDrag : Boolean read FDrag write SetDrag default True;
   property FaceColor : TColor read fColor write SetColor;
   property TickColor : TColor read fTickColor write SetTickColor;
   property Position : Longint read fCurValue write SetCurValue;
   property RotationEffect:Boolean read fRotationEffect write fRotationEffect;
   property Enabled;
   property MarkStyle:TMarkStyle read fMarkStyle write SetMarkStyle;
   property Max : Longint read fMaxValue write SetMaxValue;
   property Min : Longint read fMinValue write FMinvalue;
   property OnChange : TKnobChangeEvent read fKnobChangeEvent write fKnobChangeEvent;
   property ParentColor;
   property ParentShowHint;
   property ShowHint;
   property Visible;

end;

procedure Register;

implementation

constructor TmKnob.Create(AOwner : TComponent);
begin
   inherited Create(AOwner);
   ControlStyle := ControlStyle + [csOpaque];
   Width := 60;
   Height := 60;
   fMaxValue := 100;
   fMinValue := 0;
   fCurValue := 0;
   fRotationEffect:=false;
   fMarkStyle:=msLine;
   fTickColor:=ClBlack;
   FollowMouse:=false;
   SetColor($00B5CCBD);
   SetDrag(True);
end;

{ This function constrains a value between two other values }
function TmKnob.ConstrainValue(Minimum, Maximum, Value : Longint) : Longint;
begin
   Result := Value;
   if Value < Minimum then Result := Minimum;
   if Value > Maximum then Result := Maximum;
end;

procedure TmKnob.SetMaxValue(Value : Longint);
begin
   if Value <> fMaxValue then
       begin
          fMaxValue := Value;
          Refresh;
       end;
end;

procedure TmKnob.SetMarkStyle(Value : TMarkStyle);
begin
   if Value <> fMarkStyle then
       begin
          fMarkStyle := Value;
          Refresh;
       end;
end;

procedure TmKnob.SetCurValue(Value : Longint);
begin
   if Value <> FCurValue then
      begin
         FCurValue := ConstrainValue(fMinValue, fMaxValue, Value);
         Refresh;
         KnobChange;
      end;
end;

procedure TmKnob.SetTickColor(Value : TColor);
begin
   if Value <> fTickColor then
      begin
         fTickColor := Value;
         Refresh;
      end;

end;


procedure TmKnob.SetColor(Color : TColor);
begin
   FColor := Color;
   Refresh;
end;

procedure TmKnob.SetDrag(Value : Boolean);
begin
   if Value <> FDrag then
   begin
   	FDrag := Value;
   	Refresh;
   end;
end;

procedure TmKnob.KnobChange;
begin
	if Assigned(fKnobChangeEvent) then
   	fKnobChangeEvent(Self, FCurValue);
end;

procedure TmKnob.Paint;
var	R             : TRect;
       	Bm            : TBitMap;
        Co,Si,Angle   : Double;
        X, Y,W, H     : Integer;
        dx,dy,gx,gy   : Integer;
        OuterPoint    : TPoint;
begin
	{ Initialize offscreen BitMap }
   Bm := TBitMap.Create;
   if ( csdesigning in componentstate) then
      if Height < Width then
         Height:=Width
      else
         Width:=Height;

   Bm.Width := Width;
   Bm.Height := Height;
   Bm.Canvas.Brush.Color := clBTNFACE; //clWindow;
   R.Left := 0;
   R.Top := 0;
   R.Right := Width ;//- 1;
   R.Bottom := Height;// - 1;
   W := R.Right - R.Left -4;
   H := R.Bottom - R.Top -4;
   { This weird thing make knob "shake", emulating a rotation effect.
      Not so pretty, but I like it..............}
   if fRotationEffect then
      if (position mod 2) <> 0 then
         inc(h);

   Bm.Canvas.FillRect(R);

   with Bm.Canvas do
    begin

    Brush.Color := FaceColor;
    Pen.Width := 2;
    Pen.Color := Cl3dlight;
    ellipse(1, 1, W-1, H-1);
    Pen.Color := Clbtnshadow;
    ellipse(3, 3, W+2, H+2);

    Pen.Color := Clbtnface;
    Pen.Width := 1;
    RoundRect(2,2,w,h,w,h);
    Pen.Width := 3;
    Pen.Color := TickColor;

    if Position >= 0 then
    begin
      Brush.Color := FaceColor;
      X := W div 2;
      Y := H div 2;
      dX := W div 6;
      dY := H div 6;
      gX := W div 32;
      gY := H div 32;

      Angle:=(Position - (Min + Max)/2 ) / (Max - Min) * 5 ;

      Si:=Sin(Angle);
      Co:=Cos(Angle);
      OuterPoint.X:=Round(X + Si * (X-dx));
      OuterPoint.Y:=Round(Y - Co * (Y-dy));
      MoveTo(OuterPoint.X,OuterPoint.y);

      case MarkStyle of
        msLine   : LineTo(Round(X + Si * (X-gx)),Round(Y - Co * (Y-gy)));
  { this implementation of circle style is very poor but for my needing is enough}
        msCircle : begin
                   Brush.Color := TickColor;
                   RoundRect(OuterPoint.X-3, OuterPoint.Y-3,
                             OuterPoint.X+3, OuterPoint.Y+3,
                             OuterPoint.X+3, OuterPoint.Y+3);
                   end;
      end;
    end;
   end;

   Canvas.CopyMode := cmSrcCopy;
   Canvas.Draw(0, 0, Bm);
   bm.Destroy;
end;

procedure TmKnob.UpdatePosition(X, Y: Integer);
var
  CX,CY   : integer;
  R       : double;
  Angle   : double;

begin
     CX:=Width div 2;
     CY:=Height div 2;
     R:=Round(sqrt(sqr(CX-X)+sqr(CY-Y))) ;
     if R=0 then R:=0.0001;

     if Y < CY then
        Angle:=ArcSin((X-CX)/R)
     else
        begin
           Angle:=arcsin((CX-X)/R);
           if X > CX then
              Angle:=Angle + Pi
           else
              Angle:=Angle - Pi;
        end;
     Position:= Round(Angle * (Max - Min) / 5 + (Min + Max) / 2);
     Refresh;
end;

procedure TmKnob.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   inherited MouseDown(Button, Shift, X, Y);
   if FDrag = True then
      begin
   	FollowMouse := True;
        UpdatePosition(X,Y);
   	Refresh;
      end;
end;

procedure TmKnob.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
    inherited MouseUp(Button, Shift, X, Y);
    FollowMouse := False;
end;

procedure TmKnob.MouseMove(Shift: TShiftState; X, Y: Integer);

begin
   inherited MouseMove(Shift, X, Y);
   if FollowMouse = True then
      begin
        UpdatePosition(X,Y)
      end;
end;

procedure Register;
begin
	RegisterComponents('MCaselli', [TmKnob]);
end;

end.
